import sys

from .puzzle_data_encryption import decrypt_data


def run_puzzle(in_puzzle, in_decrypt, get_answer):
    print(in_puzzle["str"])
    if "rest" in in_puzzle:
        this_pass = get_answer()
        new_puzzle = decrypt_data(in_puzzle["rest"], this_pass, in_decrypt)
        if new_puzzle is None:
            print("This is a wrong answer. Try again!")
            sys.exit(1)
        else:
            run_puzzle(new_puzzle, in_decrypt, get_answer)



from .simple import simple
from .simple import spiced


def get_configurator(**kwargs):
    encryption = kwargs.get("encryption", "spiced")
    configurators = {
        "simple": simple.Simple,
        "spiced": spiced.Spiced,
    }
    return configurators[encryption](**kwargs)



import json
import typing


def encrypt_data(in_data, in_encrypt: typing.Callable[[str, str], str]):
    if list(in_data.keys()) == ["str"]:
        return {"str": in_data["str"]}
    rest: str = json.dumps(encrypt_data(in_data["rest"], in_encrypt))
    encrypted: str = in_encrypt(rest, in_data["pass"])
    return {"str": in_data["str"], "rest": encrypted}


def decrypt_data(
    in_rest: str, in_pass: str, in_decrypt: typing.Callable[[str, str], str | None]
):
    rest_str = in_decrypt(in_rest, in_pass)
    if rest_str is None:
        return None
    return json.loads(rest_str)



import typing

from ...encryption_algorithms import bytestr_utils as bu
from ...encryption_algorithms.simple import common
from ...encryption_algorithms.simple import simple as se
from ...puzzle_data_encryption import decrypt_data
from .. import common as cc
from .common import MODULES
from ...run_puzzle import run_puzzle


class Simple:
    def __init__(self, **kwargs):
        self._proc_hasher = kwargs.get("proc_hasher", cc.DefaultHasher)
        self._signature_hasher = kwargs.get("signature_hasher", cc.DefaultHasher)

    def get_modules(self) -> typing.List[str]:
        return MODULES

    def get_encrypt(self):
        return se.get_encrypt(self._proc_hasher, self._signature_hasher)

    def get_needed_objects(self):
        return [
            common.hash_bytes,
            common.int_to_bytes,
            common.split_encrypted_and_signature,
            common.proc_bytes,
            bu.bytestr_to_bytes,
            se.get_decrypt,
            decrypt_data,
            run_puzzle,
        ]

    def get_constants_str(
        self,
    ) -> str:
        return (
            "_DECRYPT = get_decrypt("
            f"{cc.get_hasher_name(self._proc_hasher)}, "
            f"{cc.get_hasher_name(self._signature_hasher)})"
        )



import typing
import secrets

from .common import (
    proc_bytes,
    hash_bytes,
    merge_encrypted_and_signature,
    split_encrypted_and_signature,
)
from ..bytestr_utils import bytes_to_bytestr, bytestr_to_bytes


def get_encrypt(
    proc_hasher,
    signature_hasher,
    proc_spices: typing.List[bytes],
    signature_spices: typing.List[bytes],
) -> typing.Callable[[str, str], str]:
    assert proc_spices
    assert signature_spices

    def _encrypt_bytes(in_bytes: bytes, in_pass: bytes) -> bytes:
        proc_spice = secrets.choice(proc_spices)
        signature_spice = secrets.choice(signature_spices)
        encrypted = proc_bytes(in_bytes, in_pass + proc_spice, proc_hasher)
        signature = hash_bytes(in_bytes + signature_spice, signature_hasher)
        return merge_encrypted_and_signature(encrypted, signature)

    def _encrypt(in_str: str, in_pass: str) -> str:
        encrypted_bytes = _encrypt_bytes(in_str.encode(), in_pass.encode())
        return bytes_to_bytestr(encrypted_bytes)

    return _encrypt


def get_decrypt(
    proc_hasher,
    signature_hasher,
    proc_spices: typing.List[bytes],
    signature_spices: typing.List[bytes],
) -> typing.Callable[[str, str], str | None]:
    assert proc_spices
    assert signature_spices

    def _decrypt_bytes(in_bytes: bytes, in_pass: bytes) -> bytes | None:
        for proc_spice in proc_spices:
            encrypted, signature = split_encrypted_and_signature(
                in_bytes, signature_hasher().digest_size
            )
            res = proc_bytes(encrypted, in_pass + proc_spice, proc_hasher)

            if any(
                hash_bytes(res + _, signature_hasher) == signature
                for _ in signature_spices
            ):
                return res
        return None

    def _decrypt(in_str: str, in_pass: str) -> str | None:
        res = _decrypt_bytes(bytestr_to_bytes(in_str), in_pass.encode())
        if res is not None:
            return res.decode()
        return res

    return _decrypt



import itertools
import typing


def hash_bytes(in_bytes: bytes, in_hasher) -> bytes:
    return in_hasher(in_bytes).digest()


def int_to_bytes(in_val: int) -> bytes:
    return in_val.to_bytes((in_val.bit_length() + 7) // 8, "big")


def proc_bytes(in_bytes: bytes, in_key: bytes, in_hasher) -> bytes:
    """xors the in_bytes with a sequence of bytes generated with in_key"""
    key_bytes = itertools.chain.from_iterable(
        in_hasher(in_key + int_to_bytes(block_num)).digest()
        for block_num in itertools.count(0)
    )
    return bytes(_d ^ _k for (_d, _k) in zip(in_bytes, key_bytes))


def merge_encrypted_and_signature(in_encrypted: bytes, in_signature: bytes) -> bytes:
    return in_encrypted + in_signature


def split_encrypted_and_signature(
    in_bytes: bytes, signature_size: int
) -> typing.Tuple[bytes, bytes]:
    assert len(in_bytes) >= signature_size
    encrypted = in_bytes[:-signature_size]
    signature = in_bytes[-signature_size:]
    return encrypted, signature



import inspect

from .puzzle_data_encryption import encrypt_data
from .configurators import configurators


def _create_str(in_encrypted_puzzle, configurator) -> str:
    advertisement = """# generated with puzzle-generator
#
# https://pypi.org/project/puzzle-generator/
# https://github.com/vil02/puzzle_generator/
"""
    modules: str = "\n".join("import " + _ for _ in configurator.get_modules()) + "\n"
    objects: str = "\n".join(
        inspect.getsource(_) for _ in configurator.get_needed_objects()
    )
    puzzle_data: str = f"_PUZZLE = {in_encrypted_puzzle}"
    call: str = "run_puzzle(_PUZZLE, _DECRYPT, input)"

    return (
        "\n".join(
            [
                advertisement,
                modules,
                objects,
                puzzle_data,
                configurator.get_constants_str(),
                call,
            ]
        )
        + "\n"
    )


def create(in_puzzle, **kwargs) -> str:
    configurator = configurators.get_configurator(**kwargs)
    encrypted_puzzle = encrypt_data(in_puzzle, configurator.get_encrypt())
    return _create_str(encrypted_puzzle, configurator)



import secrets
import typing

from ...encryption_algorithms import bytestr_utils as bu
from ...encryption_algorithms.simple import common
from ...encryption_algorithms.simple import spiced as sse
from ...puzzle_data_encryption import decrypt_data
from .. import common as cc
from .common import MODULES
from ...run_puzzle import run_puzzle


def _get_some_spices():
    return [secrets.token_bytes(3) for _ in range(20)]


def _list_of_bytes_to_str(in_list: typing.List[bytes]) -> str:
    return str([bu.bytes_to_bytestr(_) for _ in in_list])


def _list_of_bytes_to_codestr(in_list: typing.List[bytes]) -> str:
    raw: str = _list_of_bytes_to_str(in_list)
    return f"[bytestr_to_bytes(_) for _ in {raw}]"


class Spiced:
    def __init__(self, **kwargs):
        self._proc_hasher = kwargs.get("proc_hasher", cc.DefaultHasher)
        self._signature_hasher = kwargs.get("signature_hasher", cc.DefaultHasher)
        self._proc_spices = kwargs.get("proc_spices", _get_some_spices())
        self._signature_spices = kwargs.get("signature_spices", _get_some_spices())

    def get_modules(self) -> typing.List[str]:
        return MODULES

    def get_encrypt(self):
        return sse.get_encrypt(
            self._proc_hasher,
            self._signature_hasher,
            self._proc_spices,
            self._signature_spices,
        )

    def get_needed_objects(self):
        return [
            common.hash_bytes,
            common.int_to_bytes,
            common.split_encrypted_and_signature,
            common.proc_bytes,
            bu.bytestr_to_bytes,
            sse.get_decrypt,
            decrypt_data,
            run_puzzle,
        ]

    def get_constants_str(
        self,
    ) -> str:
        proc_spices: str = (
            f"_PROC_SPICES = {_list_of_bytes_to_codestr(self._proc_spices)}"
        )
        signature_spices: str = (
            f"_SIGNATURE_SPICES = {_list_of_bytes_to_codestr(self._signature_spices)}"
        )
        decrypt: str = (
            "_DECRYPT = get_decrypt("
            f"{cc.get_hasher_name(self._proc_hasher)}, "
            f"{cc.get_hasher_name(self._signature_hasher)}, "
            "_PROC_SPICES, "
            "_SIGNATURE_SPICES)"
        )
        return "\n".join([proc_spices, signature_spices, decrypt])
